/*
    Copyright (C) 2005 Andreas Sandberg <andreas@sandberg.pp.se>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "oss2jack.h"
#include <sys/errno.h>

#define MIXER_DEVMASK ((1 << SOUND_MIXER_VOLUME) | \
		       (1 << SOUND_MIXER_PCM)    | \
                       (1 << SOUND_MIXER_MIC))
#define MIXER_RECMASK (1 << SOUND_MIXER_MIC)
#define MIXER_STEREODEVS MIXER_DEVMASK
#define MIXER_CAPS (0)

int mixer_set_volume(int* volumep, int oss_volume)
{
	volumep[0] = oss_volume & 0xFF;
	volumep[1] = (oss_volume >> 8) & 0xFF;

	if(volumep[0] > 100)
		volumep[0] = 100;
	
	if(volumep[1] > 100)
		volumep[1] = 100;
	
	return mixer_get_volume(volumep);
}

int mixer_get_volume(int* volumep) {
	return volumep[0] | 
		(volumep[1] << 8);
}

int do_mixer_open(struct fusd_file_info* file)
{
	pprintf(DEBUG_LEVEL_INFO, "Mixer device opened...\n");
	struct mixer_instance* mixer = (struct mixer_instance*) malloc(sizeof(struct mixer_instance));
	struct device_info* info = (struct device_info*) file->device_info;
	struct oss_instance* instance;
	
	mixer->volume[0] = 100;
	mixer->volume[1] = 100;
	
	pthread_mutex_lock(&info->mutex);
	instance = oss_find_pid(info, file->pid);
	if(instance == NULL) {
		instance = oss_create(info, file->pid, NULL, mixer);
	} else {
		instance->mixer = mixer;
	}
	pthread_mutex_unlock(&info->mutex);
	
	file->private_data = instance;
	return 0;
}

int do_mixer_close(struct fusd_file_info* file)
{
	pprintf(DEBUG_LEVEL_INFO, "Mixer device closed...\n");
	struct device_info* info = (struct device_info*) file->device_info;
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct mixer_instance* mixer = instance->mixer;
	
	pthread_mutex_lock(&info->mutex);
	
	instance->mixer = NULL;
	if(instance->dsp == NULL)
		oss_delete(info, instance);
	
	pthread_mutex_unlock(&info->mutex);
	
	return 0;
}

int do_mixer_ioctl(struct fusd_file_info *file, int request, void *data)
{
	struct device_info* dinfo = (struct device_info*) file->device_info;
	struct oss_instance* instance = (struct oss_instance*) file->private_data;
	struct mixer_instance* mixer = instance->mixer;
	
	pthread_mutex_lock(&dinfo->mutex);
	int return_value = EINVAL;
	
	switch(request)
	{
	case OSS_GETVERSION:
		pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: OSS_GETVERSION\n");
		*(int*)data = SOUND_VERSION;
		return_value = 0;
		break;
		
	case SOUND_MIXER_READ_DEVMASK:
		pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SOUND_MIXER_READ_DEVMASK\n");
		*(int*)data = MIXER_DEVMASK;
		return_value = 0;
		break;
		
	case SOUND_MIXER_READ_RECMASK:
		pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SOUND_MIXER_READ_RECMASK\n");
		*(int*)data = MIXER_RECMASK;
		return_value = 0;
		break;

	case SOUND_MIXER_READ_RECSRC:
		pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SOUND_MIXER_READ_RECSRC\n");
		*(int*)data = MIXER_RECMASK;
		return_value = 0;
		break;
		
	case SOUND_MIXER_READ_STEREODEVS:
		pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SOUND_MIXER_READ_STEREODEVS\n");
		*(int*)data = MIXER_STEREODEVS;
		return_value = 0;
		break;
		
	case SOUND_MIXER_READ_CAPS:
		pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SOUND_MIXER_READ_CAPS\n");
		*(int*)data = MIXER_CAPS;
		return_value = 0;
		break;
		
	case SOUND_MIXER_WRITE_RECSRC:
		pprintf(DEBUG_LEVEL_WARNING, "IOCTL: SOUND_MIXER_WRITE_RECSRC, NOT IMPLEMENTED\n");
		break;
		
	case SOUND_MIXER_INFO:
		pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SOUND_MIXER_INFO\n");
		mixer_info* mi = (mixer_info*)data;
		
		strcpy(mi->id, "oss2jack");
		strcpy(mi->name, "OSS2Jack Emulated device");
		mi->modify_counter = dinfo->mix_modcount;
		
		return_value = 0;
		break;
		
	case SOUND_MIXER_READ_PCM:
		pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SOUND_MIXER_READ_PCM\n");
		*(int*)data = mixer_get_volume(mixer->volume);
		
		return_value = 0;
		break;
		
	case SOUND_MIXER_READ_VOLUME:
		pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SOUND_MIXER_READ_VOLUME\n");
		*(int*)data = mixer_get_volume(dinfo->volume);
		
		return_value = 0;
		break;

	case SOUND_MIXER_READ_MIC:
		pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SOUND_MIXER_READ_MIC\n");
		*(int*)data = mixer_get_volume(dinfo->rec_volume);
		
		return_value = 0;
		break;

	case SOUND_MIXER_WRITE_PCM:
		pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SOUND_MIXER_WRITE_PCM\n");
		
		*(int*)data = mixer_set_volume(dinfo->volume, *(int*)data);
		pprintf(DEBUG_LEVEL_DEBUG2, "    Setting vol: %i %i\n", dinfo->volume[0], dinfo->volume[1]);
		
		dinfo->mix_modcount++;
		return_value = 0;
		break;
		
	case SOUND_MIXER_WRITE_VOLUME:
		pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SOUND_MIXER_WRITE_VOLUME\n");
		
		*(int*)data = mixer_set_volume(dinfo->volume, *(int*)data);
		pprintf(DEBUG_LEVEL_DEBUG2, "    Setting vol: %i %i\n", dinfo->volume[0], dinfo->volume[1]);
		
		dinfo->mix_modcount++;
		return_value = 0;
		break;

	case SOUND_MIXER_WRITE_MIC:
		pprintf(DEBUG_LEVEL_DEBUG2, "IOCTL: SOUND_MIXER_WRITE_MIC\n");
		
		*(int*)data = mixer_set_volume(dinfo->rec_volume, *(int*)data);
		pprintf(DEBUG_LEVEL_DEBUG2, "    Setting vol: %i %i\n", dinfo->rec_volume[0], dinfo->rec_volume[1]);
		
		dinfo->mix_modcount++;
		return_value = 0;
		break;
		
	default:
		pprintf(DEBUG_LEVEL_WARNING, "MIXER IOCTL: Unknown IOCTL %d,  NOT IMPLEMENTED\n", request);
		break;
	}
	
	pthread_mutex_unlock(&dinfo->mutex);
	
	return return_value;
}

struct fusd_file_operations mixer_file_ops = 
{
	open: do_mixer_open,
	close: do_mixer_close,
	ioctl: do_mixer_ioctl,
};
