/*
    Copyright (C) 2004 Kor Nielsen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "oss2jack.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <syslog.h>

#include "fcntl.h"

#include <sys/mman.h>
#include <sys/errno.h>


// Global variables which define debuglevel (how much data to output) or wether
// we are daemonized, so we know where to log
int debuglevel = DEBUG_LEVEL_WARNING;
int daemonized = 0;

/**
 * Print formatted message to stdout/stderr or syslog depending on
 * importance of output and wether we are running as daemon
 * @param level
 *     Importance level, should be one of the DEBUG_LEVEL_* contants
 * @param format
 *     Message formatting rules (see printf for details)
 * @param ...
 *     Values to be formatted into message.
 */
inline void pprintf(const int level, const char *format, ...)
{
	if (level <= debuglevel)
	{
		va_list ap;
		va_start(ap, format);
		if (daemonized)
		{
			int dl;
			switch (level) {
				case DEBUG_LEVEL_CRITICAL:
					dl = LOG_CRIT;
					break;
				case DEBUG_LEVEL_ERROR:
					dl = LOG_ERR;
					break;
				case DEBUG_LEVEL_WARNING:
					dl = LOG_WARNING;
					break;
				case DEBUG_LEVEL_NOTICE:
					dl = LOG_NOTICE;
					break;
				case DEBUG_LEVEL_INFO:
					dl = LOG_INFO;
					break;
				default:
					dl = LOG_DEBUG;
			}
			vsyslog(dl, format, ap);
		}
		else
			if (level > DEBUG_LEVEL_WARNING)
				vprintf(format, ap);
			else
				vfprintf(stderr, format, ap);
		va_end(ap);
	}
}

/**
 * Print formatted message to stdout/stderr or syslog depending on
 * importance of output and wether we are running as daemon.
 * Error number (errno) will be appended as human-readable string.
 * @param level
 *     Importance level, should be one of the DEBUG_LEVEL_* contants
 * @param format
 *     Message formatting rules (see printf for details)
 * @param ...
 *     Values to be formatted into message.
 */
inline void pperror(const int level, const char *format, ...)
{
	int err = errno;
	if (level <= debuglevel)
	{
		va_list ap;
		va_start(ap, format);
		if (daemonized)
		{
			int dl;
			switch (level) {
				case DEBUG_LEVEL_CRITICAL:
					dl = LOG_CRIT;
					break;
				case DEBUG_LEVEL_ERROR:
					dl = LOG_ERR;
					break;
				case DEBUG_LEVEL_WARNING:
					dl = LOG_WARNING;
					break;
				case DEBUG_LEVEL_NOTICE:
					dl = LOG_NOTICE;
					break;
				case DEBUG_LEVEL_INFO:
					dl = LOG_INFO;
					break;
				default:
					dl = LOG_DEBUG;
			}
			vsyslog(dl, format, ap);
			vsyslog(dl, ": %s", strerror(err));
		}
		else
			if (level > DEBUG_LEVEL_WARNING)
			{
				vprintf(format, ap);
				printf(": %s\n", strerror(err));
			}
			else
			{
				vfprintf(stderr, format, ap);
				fprintf(stderr, ": %s\n", strerror(err));
			}
		va_end(ap);
	}
}

void jack_buffer_clear(jack_nframes_t nframes, float* left, float* right)
{
	float* left_end = left + nframes;
	float* l = left;
	float* r = right;
	
	while(l < left_end)
	{
		
		*l++ = 0;
		*r++ = 0;
	}
}

void jack_buffer_set_vol(jack_nframes_t nframes, float* left, float* right, 
			 float vol_l, float vol_r)
{
	float* left_end = left + nframes;
	float* l = left;
	float* r = right;
	
	while(l < left_end)
	{
		
		*l++ *= vol_l;
		*r++ *= vol_r;
	}
}



/**
 * Jackd Callback -- FIXME
 * @param nframes FIXME
 * @param arg FIXME
 * @return FIXME
 */
int jack_process (jack_nframes_t nframes, void *arg)
{
	struct device_info* info = (struct device_info*) arg;
	
	info->last_frame_size = nframes;
	
	float* out_left = (float*) jack_port_get_buffer (info->output_ports[0], nframes);
	float* out_right = (float*) jack_port_get_buffer (info->output_ports[1], nframes);
	
	float* in_left = (float*) jack_port_get_buffer (info->input_ports[0], nframes);
	float* in_right = (float*) jack_port_get_buffer (info->input_ports[1], nframes);

	// clear jack buffers
	jack_buffer_clear(nframes, out_left, out_right);
	
	pthread_mutex_lock(&info->mutex);
	
	struct oss_instance* instance = info->instances;
	while(instance != NULL)
	{
		struct dsp_instance* dsp = instance->dsp;
		float vol_l, vol_r;
		
		if(dsp == NULL || (!dsp->out_enabled))
		{
			instance = instance->next;
			continue;
		}
		
		if(instance->mixer != NULL) {
			vol_l = (float)instance->mixer->volume[0] / 100.0f;
			vol_r = (float)instance->mixer->volume[1] / 100.0f;
		} else {
			vol_l = 1.0f;
			vol_r = 1.0f;
		}

		int count = cb_get_count(&dsp->out);
		
		short* s = dsp->out.rptr;
		short* end = dsp->out.buffer + dsp->out.buffer_size + 1;
		float* out_left_end = out_left + nframes;
		float* out_l = out_left;
		float* out_r = out_right;
		
		if(dsp->out_rate_converter != NULL && cb_get_count(&dsp->out) > 0)
		{
			int err;
			float in[8192];
			float out[8192];
			float* s;
			SRC_DATA data;
			
			data.end_of_input = 0;
			data.src_ratio = dsp->out_rate_ratio;
			
			data.data_in = in;
			data.data_out = out;
			data.input_frames = cb_read(&dsp->out, in, 8192) / dsp->num_channels;
			data.output_frames = nframes;
			
			//if(data.output_frames > nframes)
			//	data.output_frames = nframes;
			
			err = src_process(dsp->out_rate_converter, &data);
			
			if(err)
			{
				printf("error: %s\n", src_strerror(err));
				break;
			}
			
			s = data.data_out;
			
			out_left_end = out_l + data.output_frames_gen;
			cb_skip(&dsp->out, data.input_frames_used * dsp->num_channels);

			if(dsp->num_channels == 2)
			{
				while(out_l < out_left_end)
				{
					*out_l++ += *s++ * vol_l;
					*out_r++ += *s++ * vol_r;
				}
			}
			
			if(dsp->num_channels == 1)
			{
				while(out_l < out_left_end)
				{
					*out_l++ += *s * vol_l;
					*out_r++ += *s++ * vol_r;
				}
			}
			dsp->out_total_samples_mixed +=  data.input_frames_used * dsp->num_channels;
		}
		else
		{
			if(count > nframes)
				count = nframes;
				
			out_left_end = out_left + count;
			
			if(dsp->num_channels == 2)
			{
				cb_read_stereo_mix(&dsp->out, out_left, out_right, vol_l, vol_r, nframes);
			}
			
			if(dsp->num_channels == 1)
			{
				cb_read_mono_mix(&dsp->out, out_left, out_right, vol_l, vol_r, nframes);
			}
			dsp->out_total_samples_mixed += count * dsp->num_channels;
		}
		
		instance = instance->next;
	}	
	
	info->samples_sent += nframes * 2;
	
	instance = info->instances;
	while(instance != NULL)
	{
		struct dsp_instance* dsp = instance->dsp;
		if(dsp == NULL)
		{
			instance = instance->next;
			continue;
		}
		
		if(dsp->out_sample_ptr != NULL)
		{
			if(dsp->out_samples_left_to_mix > 0)
			{
				int samples_written = cb_write_short(&dsp->out, dsp->out_sample_ptr,
								     dsp->out_samples_left_to_mix);
				
				dsp->out_sample_ptr += samples_written;				
				dsp->out_samples_left_to_mix -= samples_written;
			}
			
			if(dsp->out_samples_left_to_mix == 0 && cb_get_available(&dsp->out) >= dsp->frag_size / 2)
			{
				dsp->out_sample_ptr = NULL;
				pprintf(DEBUG_LEVEL_DEBUG2, "fusd_return: %d bytes\n", dsp->out_write_size);
				fusd_return(dsp->out_file_info, dsp->out_write_size);
				dsp->out_file_info = NULL;
				dsp->out_write_size = 0;
			}
		}
		do_dsp_notify_complete(instance, info);
		instance = instance->next;
	}
	
	// Process Input
	

	instance = info->instances;
	while(instance != NULL)
	{
		struct dsp_instance* dsp = instance->dsp;
		float vol_l = (float)info->rec_volume[0] / 100.0f;
		float vol_r = (float)info->rec_volume[1] / 100.0f;
		
		if(dsp == NULL || !dsp->in_enabled)
		{
			instance = instance->next;
			continue;
		}
		
		int space_available = cb_get_available(&dsp->in);
		int space_needed = nframes * dsp->num_channels;
		int samples_written = 0;
		
		if(dsp->in_rate_converter != NULL)
		{
			int err;
			float in[8192];
			float out[8192];
			SRC_DATA data;
			
			float* d = &in[0];
			float* in_left_end = in_left + nframes;
			
			if(dsp->num_channels == 2)
			{
				while(in_left < in_left_end)
				{
					*d++ = *in_left++ * vol_l;
					*d++ = *in_right++ * vol_r;
				}
			}
			
			if(dsp->num_channels == 1)
			{
				while(in_left < in_left_end)
				{
					*d++ = (*in_left++) * vol_l +
					       (*in_right++) * vol_r;
				}
			}
			
			data.end_of_input = 0;
			data.src_ratio = dsp->in_rate_ratio;
			
			data.data_in = in;
			data.data_out = out;
			data.input_frames = nframes;
			data.output_frames = 8192 / dsp->num_channels;
			
			//if(data.output_frames > nframes)
			//	data.output_frames = nframes;
			
			err = src_process(dsp->in_rate_converter, &data);
			
			if(err)
			{
				printf("error: %s\n", src_strerror(err));
				break;
			}
			
			space_needed = data.output_frames_gen;
			
			if(space_needed > space_available)
				cb_skip(&dsp->in, space_needed - space_available);
			
			samples_written = cb_write(&dsp->in, out, data.output_frames_gen * dsp->num_channels);
		}
		else
		{
			if(space_needed > space_available)
				cb_skip(&dsp->in, space_needed - space_available);
			
			if(dsp->num_channels == 1)
			{
				samples_written = cb_write_mono(&dsp->in, in_left, in_right, vol_l, vol_r, nframes);
			}
			if(dsp->num_channels == 2)
			{
				samples_written = cb_write_stereo(&dsp->in, in_left, in_right, vol_l, vol_r, nframes);
			}
		}
		
		dsp->in_total_samples_read += samples_written;
		
		if(dsp->in_sample_ptr != NULL)
		{
			int samples_read = cb_read_short(&dsp->in, dsp->in_sample_ptr, dsp->in_samples_left_to_read);
			dsp->in_samples_left_to_read -= samples_read;
			dsp->in_sample_ptr += samples_read;
			
			if(dsp->in_samples_left_to_read == 0)
			{
				dsp->in_sample_ptr = NULL;
				fusd_return(dsp->in_file_info, dsp->in_read_size);
				dsp->in_file_info = NULL;
				dsp->in_read_size = 0;
			}
		}
		
		instance = instance->next;
	}
	
	// Set master volume on output
	jack_buffer_set_vol(nframes, out_left, out_right, 
			    (float)info->volume[0] / 100.0f,
			    (float)info->volume[1] / 100.0f);
	
	pthread_mutex_unlock(&info->mutex);
	
	
	return 0;
}




/**
 * Jackd Callback -- FIXME
 * @param nframes FIXME
 * @param arg FIXME
 * @return FIXME
 */
int jack_srate (jack_nframes_t nframes, void *arg)
{
	struct device_info* info = (struct device_info*) arg;
	info->sample_rate = nframes;
	return 0;
}

/**
 * Jackd Callback -- FIXME
 * @param nframes FIXME
 * @param arg FIXME
 * @return FIXME
 */
int  jack_buffer_size (jack_nframes_t nframes, void *arg)
{
	struct device_info* info = (struct device_info*) arg;
	printf("unexpected set of jack buffer size = %d\n", nframes);
	return -1;
}

/**
 * Jackd Callback -- FIXME
 * @param desc FIXME
 */
void jack_error (const char *desc)
{
	pprintf(DEBUG_LEVEL_ERROR, "JACK error: %s\n", desc);
}

/**
 * FIXME
 */
void setscheduler(void)
{
	struct sched_param sched_param;

	if (sched_getparam(0, &sched_param) < 0) {
		pprintf(DEBUG_LEVEL_WARNING, "Scheduler getparam failed...\n");
		return;
	}
	sched_param.sched_priority = 10;
	if (!sched_setscheduler(0, SCHED_FIFO, &sched_param)) {
		pprintf(DEBUG_LEVEL_INFO, "Scheduler set to Round Robin with priority %i...\n", sched_param.sched_priority);
		return;
	}
	pprintf(DEBUG_LEVEL_WARNING, "!!!Scheduler set to Round Robin with priority %i FAILED!!!\n", sched_param.sched_priority);
}

/**
 * Displays help message. Calls exit() with value given in "result"
 * once information is printed.
 * @param version_only
 *    If this parameter is non-zero, only version will be displayed
 * @param result
 *    Return value to use when exiting at end of this function
 */
void print_help(int version_only, int result)
{
	printf(PACKAGE" version "VERSION"\n"
	       " 2004 Kor Nielsen\n");
	if (!version_only) {
		// Print usage information
		printf("\nUsage:\n"
		       "   "PACKAGE" -h|v\n"
		       "   "PACKAGE" [-d] [-q] [-V] [-n <int>]\n"
		       "\n"
		       "Options:\n"
		       "     -h   --help\n"
		       "        Display this help message.\n"
		       "     -v   --version\n"
		       "        Display program version.\n"
		       "     -d   --detach\n"
		       "        Detach from console when we have our connection\n"
		       "        to Jack server and FUSD. Once detached logging\n"
		       "        will happen to syslog using daemon facility.\n"
		       "     -n <int>   --devnum <int>\n"
		       "        Use device with number <int> instead of without\n"
		       "        any numbering. Allowed is -1 .. 99.\n"
		       "        E.g. If <int> = 5, then we will setup /dev/dsp5\n"
		       "        instead of /dev/dsp\n"
		       "     -q   --quiet\n"
		       "        Be more quiet\n"
		       "     -V   --verbose\n"
		       "        Be more verbose\n"
		       "\n"
		       "Exit values:\n"
		       "      0  Success in execution.\n"
		       "      1  Invalid commandline arguments.\n"
		       "      2  Failed to connect to jackd.\n"
		       "      3  Failed to setup our device(s) under /dev/.\n"
		       "      4  Failed to detach from console\n"
		       "      5  Termination because we lost connection to jackd\n");
	}
	exit(result);
}

/**
 * Parses commandline arguments
 * @param argc Count of arguments
 * @param argv Array of string containing the arguments
 * @param detach OUT: will be set to 1 if we should detach from terminal
 * @param dspnum OUT: will be set to the device number to use for /dev/dsp%d
 */
void parse_arguments(int argc, char*argv[], int*detach, int*dspnum)
{
	int i = 0;
	*detach = 0; // Default, do not detach from console
	*dspnum = -1; // Default value, no device number!
	while (argv[++i])
	{
		if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0)
		{
			print_help(0, ERR_SUCCESS);
		}
		else if (strcmp(argv[i], "-v") == 0 || strcmp(argv[i], "--version") == 0)
		{
			print_help(1, ERR_SUCCESS);
		}
		else if (strcmp(argv[i], "-d") == 0 || strcmp(argv[i], "--detach") == 0)
		{
			*detach = 1;
		}
		else if (strcmp(argv[i], "-q") == 0 || strcmp(argv[i], "--quiet") == 0)
		{
			if (debuglevel > 0) debuglevel--;
		}
		else if (strcmp(argv[i], "-V") == 0 || strcmp(argv[i], "--verbose") == 0)
		{
			if (debuglevel < 99) debuglevel++;
		}
		else if (strcmp(argv[i], "-n") == 0 || strcmp(argv[i], "--devnum") == 0)
		{
			if (argv[++i] && strspn(argv[i], "0123456789") == strlen(argv[i]))
			{
				sscanf(argv[i], "%d", dspnum);
				if (*dspnum < -1 || *dspnum > MAX_DEVICE_NUMBER)
				{
					printf("Device number is out of range.\n");
					print_help(0, ERR_BADARG);
				}
			}
			else
			{
				// We got no device number or garbage
				printf("Invalid device number, expecting integer in range is -1 .. %d\n", MAX_DEVICE_NUMBER);
				print_help(0, ERR_BADARG);
			}
		}
		else
		{
			printf("Unknown option '%s'\n", argv[i]);
			print_help(0, ERR_BADARG);
		}
	} // End while
}

void jack_shutdown(void* arg)
{
	struct device_info* info = (struct device_info*) arg;
	
	printf("Jack shut us down; reopening jack in 2 seconds\n");
	info->restart_interval = 2;
}


void jack_init(struct device_info* info)
{
	const char** ports;
	
	jack_set_error_function(jack_error);
	
	info->client = jack_client_new("oss-dsp");
	
	if(info->client == NULL)
	{
		pprintf(DEBUG_LEVEL_CRITICAL, "Error registering jack client. Make sure that jackd is running.\n");
		exit(ERR_JACKD_FAIL);
	}

	info->output_ports[0] = jack_port_register(info->client, "out_left", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
	info->output_ports[1] = jack_port_register(info->client, "out_right", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
	
	info->input_ports[0] = jack_port_register(info->client, "in_left", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
	info->input_ports[1] = jack_port_register(info->client, "in_right", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
	
	
	jack_set_process_callback(info->client, jack_process, info);
	jack_set_sample_rate_callback(info->client, jack_srate, info);
	jack_set_buffer_size_callback(info->client, jack_buffer_size, info);
	
	jack_on_shutdown(info->client, jack_shutdown, info);
	
	info->last_frame_size = 64;

	jack_activate(info->client);

	ports = jack_get_ports (info->client, NULL, NULL, JackPortIsPhysical|JackPortIsInput);

	jack_connect (info->client, jack_port_name(info->output_ports[0]), ports[0]);
	jack_connect (info->client, jack_port_name(info->output_ports[1]), ports[1]);
	
	free(ports);
	
	ports = jack_get_ports (info->client, NULL, NULL, JackPortIsPhysical|JackPortIsOutput);

	jack_connect (info->client, ports[0], jack_port_name(info->input_ports[0]));
	jack_connect (info->client, ports[1], jack_port_name(info->input_ports[1]));

	free(ports);
}

void* jack_restart_thread(void* arg)
{
	printf("Hello from the restart thread\n");
	struct device_info* info = (struct device_info*) arg;
	
	while(1)
	{
		sleep(1);
		if(info->restart_interval > 0)
		{
			printf("restart_interval: %d\n", info->restart_interval);
			info->restart_interval--;
			if(info->restart_interval == 0)
			{
				const char** ports;
				int result;
				printf("Re-activating jack\n");
				//result = jack_activate(info->client);
				printf("Activated: %d\n", result);
				
				jack_client_close(info->client);
				jack_init(info);
				/*
				ports = jack_get_ports (info->client, NULL, NULL, JackPortIsPhysical|JackPortIsInput);

				jack_connect (info->client, jack_port_name(info->output_ports[0]), ports[0]);
				jack_connect (info->client, jack_port_name(info->output_ports[1]), ports[1]);
				
				free(ports);
				
				ports = jack_get_ports (info->client, NULL, NULL, JackPortIsPhysical|JackPortIsOutput);
			
				jack_connect (info->client, ports[0], jack_port_name(info->input_ports[0]));
				jack_connect (info->client, ports[1], jack_port_name(info->input_ports[1]));
			
				free(ports);*/
				//jack_init(info);
			}
		}
	}
}


/**
 * Program entry-point
 * @param argc Count of arguments
 * @param argv Array of string containing the arguments
 * @return Program return value
 */
int main(int argc, char *argv[])
{
	int dspnum, detach;
	pthread_t thread;
	pthread_attr_t thread_attributes;
	
	// Allow users to provide some options
	parse_arguments(argc, argv, &detach, &dspnum);

	mlockall(MCL_CURRENT | MCL_FUTURE);
	
	struct device_info* info;
	
	info = (struct device_info*) malloc(sizeof(struct device_info));
	
	info->restart_interval = 0;
	
	pthread_mutex_init(&info->mutex, NULL);
	
	info->instances = NULL;
	info->samples_sent = 0;
	
	info->volume[0] = 100;
	info->volume[1] = 100;
	info->rec_volume[0] = 100;
	info->rec_volume[1] = 100;
	info->mix_modcount = 0;
	
	jack_init(info);
	
	pthread_attr_init(&thread_attributes);
	pthread_create(&thread, &thread_attributes, jack_restart_thread, info);
	
	int dsp_fd = 0;
	int mixer_fd = 0;
	char dsp_device[20];
	char mixer_device[20];
	if (dspnum >= 0)
	{
		snprintf(dsp_device, sizeof(dsp_device), "/dev/sound/dsp%d", dspnum);
		snprintf(mixer_device, sizeof(mixer_device), "/dev/sound/mixer%d", dspnum);
	}
	else
	{
		strcpy(dsp_device, "/dev/sound/dsp");
		strcpy(mixer_device, "/dev/sound/mixer");
	}
	pprintf(DEBUG_LEVEL_INFO, "Creating DSP device '%s' and mixer device '%s'.\n", dsp_device, mixer_device);
#ifdef FUSD_KOR_HACKED_VERSION
	if(dsp_fd = fusd_register(dsp_device, "sound", dsp_device+11, 0666, info, &dsp_file_ops) < 0)
#else
	if(dsp_fd = fusd_register(dsp_device, 0666, info, &dsp_file_ops) < 0)
#endif
	{
		pperror(DEBUG_LEVEL_CRITICAL, "Could not create DSP device\n");
		exit(ERR_FUSD_FAIL);
	}
#ifdef FUSD_KOR_HACKED_VERSION
	if(dsp_fd = fusd_register(mixer_device, "sound", mixer_device+11, 0666, info, &mixer_file_ops) < 0)
#else
	if(dsp_fd = fusd_register(mixer_device, 0666, info, &mixer_file_ops) < 0)
#endif
	{
		fusd_unregister(dsp_fd);
		pperror(DEBUG_LEVEL_CRITICAL, "Could not create mixer device\n");
		exit(ERR_FUSD_FAIL);
	}

	if (detach)
	{
		if (daemon(0, 0) == -1)
		{
			// We failed to daemonize...
			pperror(DEBUG_LEVEL_CRITICAL, "Could not detach from terminal\n");
			exit(ERR_DETACH_FAIL);
		}
		else
		{
			openlog(PACKAGE, 0, LOG_DAEMON);
			daemonized = 1;
		}
	}
	setscheduler();

	pprintf(DEBUG_LEVEL_INFO, "Running...\n");
	fusd_run();
	
	fusd_unregister(dsp_fd);
	fusd_unregister(mixer_fd);
	pthread_mutex_destroy(&info->mutex);
	return 0;
}
